#include <stdint.h>
#include <stdbool.h>
#include <stdarg.h>
#include <intrin.h>
#include "comdbg.h"

#define COM_BASE_IO       0x3F8
#define DEFAULT_BAUD_RATE 115200

#define SER_RBR(x) ((x)+0)
#define SER_THR(x) ((x)+0)
#define SER_DLL(x) ((x)+0)
#define SER_DLM(x) ((x)+1)
#define SER_LCR(x) ((x)+3)
#define SER_MCR(x) ((x)+4)
#define SER_LSR(x) ((x)+5)
#define SR_LCR_CS8  0x03
#define SR_LCR_ST1  0x00
#define SR_LCR_PNO  0x00
#define SR_LCR_DLAB 0x80
#define SR_MCR_DTR  0x01
#define SR_MCR_RTS  0x02
#define SR_LSR_TBE  0x20

void com_init(void)
{
	// set baud rate and data format (8N1), turn on DTR and RTS
    __outbyte(SER_MCR(COM_BASE_IO), SR_MCR_DTR | SR_MCR_RTS);

	// set DLAB
    uint8_t lcr = __inbyte(SER_LCR(COM_BASE_IO)) | SR_LCR_DLAB;
    __outbyte(SER_LCR(COM_BASE_IO), lcr);

	// set baud rate
    uint32_t divisor = 115200 / DEFAULT_BAUD_RATE;
    __outbyte(SER_DLL(COM_BASE_IO), (divisor & 0xFF));
    __outbyte(SER_DLM(COM_BASE_IO), (divisor >> 8) & 0xFF);

	// reset DLAB and set 8N1 format
    __outbyte(SER_LCR(COM_BASE_IO), SR_LCR_CS8 | SR_LCR_ST1 | SR_LCR_PNO);

	// read junk out of the RBR
	__inbyte(SER_RBR(COM_BASE_IO));
}

void com_putch(char ch) {
	if (ch == '\n') com_putch('\r');
	while ((__inbyte(SER_LSR(COM_BASE_IO)) & SR_LSR_TBE) == 0);
	__outbyte(SER_THR(COM_BASE_IO), ch);
}

void com_puts(const char* msg) {
	while (*msg) com_putch(*msg++);
}

void com_printf(const char* format, ...)
{
	va_list ap;
	char    ch;
	
	va_start(ap, format);

	while ((ch = *format++) != 0)
	{
		// Print non-formatted output
		if (ch != '%') {
			com_putch(ch);
			continue;
		}

		// Read formatting parameter
		size_t fillsize = 0;
		char   fillchar = ' ';
		bool   rightfill = false;

		if (*format == '-') {
			rightfill = true;
			format++;
		}

		while ((ch = *format++) >= '0' && ch <= '9') {
			fillsize = (fillsize * 10) + (ch - '0');
			if (fillsize == 0 && !rightfill) fillchar = '0';
		}

		// Skip floating point formatting parameter
		if (ch == '.') {
			while ((ch = *format++) >= '0' && ch <= '9');
		}

		// skip long format specifier
		if (ch == 'l') {
			if ((ch = *format++) == 0) break;
		}

		switch (ch)
		{
			case 'c': {
				if (!rightfill) while (fillsize-- > 1) com_putch(fillchar);
				com_putch(va_arg(ap, int) & 0xFF);
				if (rightfill) while (fillsize-- > 1) com_putch(fillchar);
			}
			break;
			case 's': {
				const char* strarg = ((strarg = va_arg(ap, const char*)) == NULL ? "(null)" : strarg);
				size_t      strlen = 0;

				for (; strarg[strlen]; strlen++);
				if (!rightfill) for (; strlen < fillsize; strlen++) com_putch(fillchar);
				while (*strarg) com_putch(*strarg++);
				if (rightfill) for (; strlen < fillsize; strlen++) com_putch(fillchar);
			}
			break;
			case 'w':
				if (*format != 's' && *format != 'S') continue;
				format++; //-V796
			case 'S': {
				const wchar_t* strarg = ((strarg = va_arg(ap, const wchar_t*)) == NULL ? L"(null)" : strarg);
				size_t         strlen = 0;

				for (; strarg[strlen]; strlen++);
				if (!rightfill) for (; strlen < fillsize; strlen++) com_putch(fillchar);
				while (*strarg) com_putch(*strarg++ & 0xFF);
				if (rightfill) for (; strlen < fillsize; strlen++) com_putch(fillchar);
			}
			break;

			case 'p':
#ifdef _M_X64
			{
				uint64_t ptr = va_arg(ap, uint64_t);
				com_printf("%08X%08X", (uint32_t)(ptr >> 32), (uint32_t)ptr);
			}
			break;
#else
				ch = 'X'; // %p same as '%08X'
				fillsize = 8;
				fillchar = '0'; //-V796
#endif
			case 'b': case 'o':
			case 'u': case 'd': case 'i':
			case 'x': case 'X':
			{
				uint32_t numb = va_arg(ap, uint32_t);
				uint8_t  base = (ch == 'b' ? 2 : (ch == 'o' ? 8 : (ch == 'x' || ch == 'X' ? 16 : 10)));

				char   numbuff[128];
				size_t numsize = 0;
				bool   negative = false;

				if ((ch == 'd' || ch == 'i') && ((int32_t)numb) < 0) {
					numb = (uint32_t)(-((int32_t)numb));
					negative = true;
				}

				for (;;) {
					uint8_t digit = numb % base;
					numbuff[numsize++] = digit < 10 ? (digit + '0') : (digit - 10 + (ch == 'x' ? 'a' : 'A'));
					if (numsize == sizeof(numbuff)) break;
					if ((numb /= base) == 0) break;
				}

				if (!rightfill) {
					if (fillsize > sizeof(numbuff)) fillsize = sizeof(numbuff);
					while (numsize + negative < fillsize) numbuff[numsize++] = fillchar;
				}

				if (negative) com_putch('-');
				for (size_t i = numsize; i > 0; i--) com_putch(numbuff[i - 1]);
				
				if (rightfill) {
					for (; numsize + negative < fillsize; numsize++) com_putch(fillchar);
				}
			}
			break;
		}
	}
	va_end(ap);
}


void com_dump_mem(const void* addr, size_t size, size_t cols)
{
	const uint8_t* ptr = (const uint8_t*)addr;
	const uint8_t* end = ptr + size;
	size_t len = 0, max = cols ? cols : 8;

	com_printf("== DUMP %u bytes %p-%p ==\n", (uint32_t)size, ptr, end);

	while (ptr < end) {
		com_printf("0x%02X", *ptr++);
		com_puts((++len < max && ptr < end) ? ", " : "\n");
		if (len == max) len = 0;
	}

	for (max = max * 6 - 2; max--; ) com_putch('=');
	com_putch('\n');
}

void com_dump_dwords(const uint32_t dwords[], size_t count, size_t cols)
{
	size_t len = 0, max = cols ? cols : 8;

	for (size_t i = 0; i < count;) {
		if (len == 0) com_printf("%p: ", &dwords[i]);
		com_printf("%08X", dwords[i++]);
		com_putch((++len < max && i < count) ? ' ' : '\n');
		if (len == max) len = 0;
	}

	max = max * 9 + sizeof(void*) * 2 + 2;
	while (max--) com_putch('=');
	com_putch('\n');
}
